//------------------------------------------------------------------------------
// blocks.cpp
//------------------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//------------------------------------------------------------------------------
// Copyright (C) 2010 "Blzut3" <admin@maniacsvault.net>
//------------------------------------------------------------------------------

#include <cstring>

#include "blocks.h"

#define READINT32(pointer) ((uint32_t((uint8_t)(*pointer))) | (uint32_t(uint8_t(*(pointer+1)))<<8) | (uint32_t(uint8_t(*(pointer+2)))<<16) | (uint32_t(uint8_t(*(pointer+3)))<<24))
#define WRITEINT32(pointer, integer) *pointer = (uint8_t)(integer&0xFF);*(pointer+1) = (uint8_t)((integer>>8)&0xFF);*(pointer+2) = (uint8_t)((integer>>16)&0xFF);*(pointer+3) = (uint8_t)((integer>>24)&0xFF);

Response::Response() : giveID(0), link(0), log(0)
{
	costID[0] = costID[1] = costID[2] = 0;
	costAmount[0] = costAmount[1] = costAmount[2] = 0;
	memset(response, 0, 32);
	memset(yesMessage, 0, 80);
	memset(noMessage, 0, 80);
}

void Response::Read(const char* buffer)
{
	giveID = READINT32(&buffer[0]);
	costID[0] = READINT32(&buffer[4]);
	costID[1] = READINT32(&buffer[8]);
	costID[2] = READINT32(&buffer[12]);
	costAmount[0] = READINT32(&buffer[16]);
	costAmount[1] = READINT32(&buffer[20]);
	costAmount[2] = READINT32(&buffer[24]);
	memcpy(response, buffer + 28, 33);
	memcpy(yesMessage, buffer + 60, 81);
	link = READINT32(&buffer[140]);
	log = READINT32(&buffer[144]);
	memcpy(noMessage, buffer + 148, 81);
}

void Response::Save(std::ostream &out) const
{
	char outBuffer[228];
	WRITEINT32(&outBuffer[0], giveID);
	WRITEINT32(&outBuffer[4], costID[0]);
	WRITEINT32(&outBuffer[8], costID[1]);
	WRITEINT32(&outBuffer[12], costID[2]);
	WRITEINT32(&outBuffer[16], costAmount[0]);
	WRITEINT32(&outBuffer[20], costAmount[1]);
	WRITEINT32(&outBuffer[24], costAmount[2]);
	memcpy(outBuffer + 28, response, 32);
	memcpy(outBuffer + 60, yesMessage, 80);
	WRITEINT32(&outBuffer[140], link);
	WRITEINT32(&outBuffer[144], log);
	memcpy(outBuffer + 148, noMessage, 80);
	out.write(outBuffer, 228);
}

Speech::Speech() : speakerID(0), dropID(0), link(0), voiceNumber(0)
{
	checkID[0] = checkID[1] = checkID[2] = 0;
	memset(name, 0, 17);
	memset(sound, 0, 9);
	memset(backdrop, 0, 9);
	memset(dialog, 0, 321);
}

void Speech::Read(const char* buffer, bool oldFormat)
{
	speakerID = READINT32(&buffer[0]);
	dropID = READINT32(&buffer[4]);
	if(!oldFormat)
	{
		checkID[0] = READINT32(&buffer[8]);
		checkID[1] = READINT32(&buffer[12]);
		checkID[2] = READINT32(&buffer[16]);
		link = READINT32(&buffer[20]);
		memcpy(name, buffer + 24, 16);
		memcpy(sound, buffer + 40, 8);
		memcpy(backdrop, buffer + 48, 8);
		memcpy(dialog, buffer + 56, 320);
	}
	else
	{
		voiceNumber = READINT32(&buffer[8]);
		memcpy(name, buffer + 12, 16);
		memcpy(dialog, buffer + 28, 320);
	}

	for(unsigned int i = 0;i < 5;i++)
		responses[i].Read(buffer + (oldFormat ? 348 : 376) + i*228);
}

void Speech::Save(std::ostream &out, bool oldFormat) const
{
	char outBuffer[376];
	WRITEINT32(&outBuffer[0], speakerID);
	WRITEINT32(&outBuffer[4], dropID);
	if(!oldFormat)
	{
		WRITEINT32(&outBuffer[8], checkID[0]);
		WRITEINT32(&outBuffer[12], checkID[1]);
		WRITEINT32(&outBuffer[16], checkID[2]);
		WRITEINT32(&outBuffer[20], link);
		memcpy(outBuffer + 24, name, 16);
		memcpy(outBuffer + 40, sound, 8);
		memcpy(outBuffer + 48, backdrop, 8);
		memcpy(outBuffer + 56, dialog, 320);
	}
	else
	{
		WRITEINT32(&outBuffer[8], voiceNumber);
		memcpy(outBuffer + 12, name, 16);
		memcpy(outBuffer + 28, dialog, 320);
	}
	out.write(outBuffer, oldFormat ? 348 : 376);

	for(unsigned int i = 0;i < 5;i++)
		responses[i].Save(out);
}
