//------------------------------------------------------------------------------
// compile.cpp
//------------------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//------------------------------------------------------------------------------
// Copyright (C) 2010 "Blzut3" <admin@maniacsvault.net>
//------------------------------------------------------------------------------

#include <cstring>
#include <cmath>
#include <iostream>
#include <list>
using namespace std;

#include "blocks.h"
#include "compile.h"
#include "main.h"
#include "scanner.h"

void IgnoreKey(Scanner &sc)
{
	Warning("%d:%d:Ignoring unknown key \"%s\".", sc.GetLine(), sc.GetLinePos(), sc.str.c_str());
	if(sc.CheckToken('{'))
	{
		int level = 1;
		while(sc.TokensLeft())
		{
			if(sc.CheckToken('}'))
			{
				level--;
				if(level == 0)
					break;
			}
			else if(sc.CheckToken('{'))
				level++;
			else
				sc.GetNextToken();
		}
	}
	else
	{
		sc.MustGetToken('=');
		sc.GetNextToken();
		sc.MustGetToken(';');
	}
}

bool GetKey(Scanner &sc, char expectedToken)
{
	if(sc.CheckToken('='))
	{
		if(sc.CheckToken(expectedToken))
			return true;
		sc.GetNextToken();
		sc.MustGetToken(';');
		return false;
	}
	else
		IgnoreKey(sc);
}

bool ParseCost(Scanner &sc, Response &response, int index)
{
	if(sc.CheckToken('{'))
	{
		unsigned int costs = 0;

		while(!sc.CheckToken('}'))
		{
			sc.MustGetToken(TK_Identifier);
			if(sc.str.compare("item") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					response.costID[index] = sc.number;
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("amount") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					response.costAmount[index] = sc.number;
					sc.MustGetToken(';');
				}
			}
			else
				IgnoreKey(sc);
		}
	}
	else
	{
		Warning("%d:%d:Cost not a sub-block as expected.", sc.GetLine(), sc.GetLinePos());
		IgnoreKey(sc);
	}
	return false;
}

bool ParseChoice(Scanner &sc, Response &response)
{
	if(sc.CheckToken('{'))
	{
		unsigned int costs = 0;
		bool displayCost = true;
		bool closeDialog = false;

		while(!sc.CheckToken('}'))
		{
			sc.MustGetToken(TK_Identifier);
			if(sc.str.compare("text") == 0)
			{
				if(GetKey(sc, TK_StringConst))
				{
					strncpy(response.response, sc.str.c_str(), 32);
					if(sc.str.length() > 32)
						Warning("%d:%d:Text too long, truncating to 32 characters.", sc.GetLine(), sc.GetLinePos());
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("cost") == 0)
			{
				if(costs > 2)
					Warning("%d:%d:Too many cost blocks, ignoring.", sc.GetLine(), sc.GetLinePos());
				else
				{
					ParseCost(sc, response, costs);
					costs++;
				}
			}
			else if(sc.str.compare("displaycost") == 0)
			{
				if(GetKey(sc, TK_BoolConst))
				{
					displayCost = sc.boolean;
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("yesmessage") == 0)
			{
				if(GetKey(sc, TK_StringConst))
				{
					strncpy(response.yesMessage, sc.str.c_str(), 80);
					if(sc.str.length() > 80)
						Warning("%d:%d:Yesmessage too long, truncating to 80 characters.", sc.GetLine(), sc.GetLinePos());
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("nomessage") == 0)
			{
				if(GetKey(sc, TK_StringConst))
				{
					strncpy(response.noMessage, sc.str.c_str(), 80);
					if(sc.str.length() > 80)
						Warning("%d:%d:Nomessage too long, truncating to 80 characters.", sc.GetLine(), sc.GetLinePos());
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("log") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					response.log = sc.number;
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("giveitem") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					response.giveID = sc.number;
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("nextpage") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					response.link = sc.number;
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("closedialog") == 0)
			{
				if(GetKey(sc, TK_BoolConst))
				{
					closeDialog = sc.boolean;
					sc.MustGetToken(';');
				}
			}
			else
				IgnoreKey(sc);
		}

		if(!displayCost)
			response.costAmount[0] *= -1;
		if(!closeDialog)
			response.link *= -1;
	}
	else
	{
		Warning("%d:%d:Choice not a sub-block as expected.", sc.GetLine(), sc.GetLinePos());
		IgnoreKey(sc);
	}
	return false;
};

bool ParseIfItem(Scanner &sc, Speech &speech, int index)
{
	if(sc.CheckToken('{'))
	{
		unsigned int items = 0;
		unsigned int choices = 0;

		while(!sc.CheckToken('}'))
		{
			sc.MustGetToken(TK_Identifier);
			if(sc.str.compare("item") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					speech.checkID[index] = sc.number;
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("page") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					if(index == 0)
						speech.link = sc.number;
					else if(sc.number != speech.link)
						Warning("%d:%d:Ignoring differing ifitem jump.", sc.GetLine(), sc.GetLinePos());
					sc.MustGetToken(';');
				}
			}
			else
				IgnoreKey(sc);
		}
	}
	else
	{
		Warning("%d:%d:Ifitem not a sub-block as expected.", sc.GetLine(), sc.GetLinePos());
		IgnoreKey(sc);
	}
	return false;
}

bool ParsePage(Scanner &sc, Speech &speech, bool oldformat)
{
	if(sc.CheckToken('{'))
	{
		unsigned int items = 0;
		unsigned int choices = 0;

		while(!sc.CheckToken('}'))
		{
			sc.MustGetToken(TK_Identifier);
			if(sc.str.compare("name") == 0)
			{
				if(GetKey(sc, TK_StringConst))
				{
					strncpy(speech.name, sc.str.c_str(), 16);
					if(sc.str.length() > 16)
						Warning("%d:%d:Character name too long, truncating to 16 characters.", sc.GetLine(), sc.GetLinePos());
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("panel") == 0)
			{
				if(GetKey(sc, TK_StringConst))
				{
					strncpy(speech.backdrop, sc.str.c_str(), 8);
					if(sc.str.length() > 8)
						Warning("%d:%d:Panel name too long, truncating to 8 characters.", sc.GetLine(), sc.GetLinePos());
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("voice") == 0)
			{
				if(oldformat)
				{
					Warning("%d:%d:Voice not applicable to old format scripts, use voicenumber instead.", sc.GetLine(), sc.GetLinePos());
					IgnoreKey(sc);
				}
				else
				{
					if(GetKey(sc, TK_StringConst))
					{
						strncpy(speech.sound, sc.str.c_str(), 8);
						if(sc.str.length() > 8)
							Warning("%d:%d:Voice name too long, truncating to 8 characters.", sc.GetLine(), sc.GetLinePos());
						sc.MustGetToken(';');
					}
				}
			}
			else if(sc.str.compare("voicenumber") == 0)
			{
				if(!oldformat)
				{
					Warning("%d:%d:Voicenumber only applicable to old format scripts, use voice instead.", sc.GetLine(), sc.GetLinePos());
					IgnoreKey(sc);
				}
				else
				{
					if(GetKey(sc, TK_IntConst))
					{
						speech.voiceNumber = sc.number;
						sc.MustGetToken(';');
					}
				}
			}
			else if(sc.str.compare("dialog") == 0)
			{
				if(GetKey(sc, TK_StringConst))
				{
					strncpy(speech.dialog, sc.str.c_str(), 320);
					if(sc.str.length() > 320)
						Warning("%d:%d:Dialog too long, truncating to 320 characters. (Honestly, did you really think this string is going to fit on screen anyways?)", sc.GetLine(), sc.GetLinePos());
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("drop") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					speech.dropID = sc.number;
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("ifitem") == 0)
			{
				if(items > 2)
					Warning("%d:%d:Too many ifitem blocks, ignoring.", sc.GetLine(), sc.GetLinePos());
				else
				{
					ParseIfItem(sc, speech, items);
					items++;
				}
			}
			else if(sc.str.compare("choice") == 0)
			{
				if(choices > 4)
					Warning("%d:%d:Too many choices blocks, ignoring.", sc.GetLine(), sc.GetLinePos());
				else
				{
					ParseChoice(sc, speech.responses[choices]);
					choices++;
				}
			}
			else
				IgnoreKey(sc);
		}
		return true;
	}
	else
	{
		Warning("%d:%d:Page not a sub-block as expected.", sc.GetLine(), sc.GetLinePos());
		IgnoreKey(sc);
	}
	return false;
}

bool ParseConversation(Scanner &sc, list<Speech> &script, bool oldformat)
{
	if(sc.CheckToken('{'))
	{
		list<Speech> newScripts;

		int actorID = -1;
		while(!sc.CheckToken('}'))
		{
			sc.MustGetToken(TK_Identifier);
			if(sc.str.compare("actor") == 0)
			{
				if(GetKey(sc, TK_IntConst))
				{
					actorID = sc.number;
					sc.MustGetToken(';');
				}
			}
			else if(sc.str.compare("page") == 0)
			{
				Speech newSpeech;
				if(ParsePage(sc, newSpeech, oldformat))
				{
					newScripts.push_back(newSpeech);
				}
			}
			else
				IgnoreKey(sc);
		}

		if(actorID == -1)
		{
			Error("%d:%d:A conversation block must contain an actor property.", sc.GetLine(), sc.GetLinePos());
		}
		else
		{
			for(list<Speech>::iterator iter = newScripts.begin();iter != newScripts.end();iter++)
			{
				(*iter).speakerID = actorID;
				script.push_back(*iter);
			}
		}
		return true;
	}
	else
	{
		Warning("%d:%d:Conversation not a sub-block as expected.", sc.GetLine(), sc.GetLinePos());
		IgnoreKey(sc);
	}
	return false;
}

void Compile(const char* buffer, int size, std::ostream &out, bool oldformat)
{
	Scanner::SetErrorCallback(Error);

	Scanner sc(buffer, size);

	list<Speech> script;

	while(sc.TokensLeft())
	{
		sc.MustGetToken(TK_Identifier);
		if(sc.str.compare("conversation") == 0)
		{
			ParseConversation(sc, script, oldformat);
		}
		else
		{
			IgnoreKey(sc);
		}
	}

	for(list<Speech>::iterator iter = script.begin();iter != script.end();iter++)
	{
		(*iter).Save(out, oldformat);
	}
}

void Decompile(const char* buffer, int size, std::ostream &out, bool oldformat)
{
	bool sourceIsOld = false;
	if(size % 1488 == 0)
		sourceIsOld = true;
	else if(size % 1516 != 0)
		Error("Compiled file unrecognizable as a Strife dialog.");

	list<Speech> script;
	for(int i = 0;i < size;i += (sourceIsOld ? 1488 : 1516))
	{
		Speech next;
		next.Read(buffer + i, sourceIsOld);
		script.push_back(next);
	}

	int lastID = -1;
	for(list<Speech>::iterator iter=script.begin();iter != script.end();iter++)
	{
		Speech &speech = *iter;
		if(lastID != speech.speakerID)
		{
			if(lastID != -1)
				out << "}\n\n";
			out << "conversation\n{\n\tactor = " << speech.speakerID << ";\n";
			lastID = speech.speakerID;
		}

		// Pages
		out << "\tpage\n\t{\n";
		if(speech.name[0] != 0)
			out << "\t\tname = \"" << Scanner::Escape(speech.name) << "\";\n";
		if(speech.backdrop[0] != 0)
			out << "\t\tpanel = \"" << Scanner::Escape(speech.backdrop) << "\";\n";
		if(speech.sound[0] != 0)
			out << "\t\tvoice = \"" << Scanner::Escape(speech.sound) << "\";\n";
		if(speech.voiceNumber != 0)
			out << "\t\tvoiceNumber = " << speech.voiceNumber << ";\n";
		if(speech.dialog[0] != 0)
			out << "\t\tdialog = \"" << Scanner::Escape(speech.dialog) << "\";\n";
		if(speech.dropID != 0)
			out << "\t\tdrop = " << speech.dropID << ";\n";
		int lastUsedCheck = 0;
		for(int i = 0;i < 3;i++)
		{
			if(speech.checkID[i] != 0)
				lastUsedCheck = i+1;
		}
		for(int i = 0;i < lastUsedCheck;i++)
		{
			out << "\t\tifitem\n\t\t{\n\t\t\titem = " << speech.checkID[i] << ";\n\t\t\tpage = " << speech.link << ";\n\t\t}\n";
		}
		for(int i = 0;i < 5;i++)
		{
			if(speech.responses[i].response[0] != 0)
			{
				out << "\t\tchoice\n\t\t{\n\t\t\ttext = \"" << Scanner::Escape(speech.responses[i].response) << "\";\n";
				bool displayCost = speech.responses[i].costAmount[0] >= 0;
				int lastUsedCost = 0;
				for(int j = 0;j < 3;j++)
				{
					if(speech.responses[i].costID[j] != 0)
						lastUsedCost = j+1;
				}
				for(int j = 0;j < 3;j++)
				{
					out << "\t\t\tcost\n\t\t\t{\n\t\t\t\titem = " << speech.responses[i].costID[j] << ";\n\t\t\t\tamount = " << static_cast<int> (abs(static_cast<double> (speech.responses[i].costAmount[j]))) << ";\n\t\t\t}\n";
				}
				if(speech.responses[i].costID[0] != 0)
					out << "\t\t\tdisplaycost = " << (displayCost ? "true" : "false") << ";\n";
				if(speech.responses[i].yesMessage[0] != 0)
					out << "\t\t\tyesmessage = \"" << Scanner::Escape(speech.responses[i].yesMessage) << "\";\n";
				if(speech.responses[i].noMessage[0] != 0)
					out << "\t\t\tnomessage = \"" << Scanner::Escape(speech.responses[i].noMessage) << "\";\n";
				if(speech.responses[i].log != 0)
					out << "\t\t\tlog = " << speech.responses[i].log << ";\n";
				if(speech.responses[i].giveID != 0)
					out << "\t\t\tgiveitem = " << speech.responses[i].giveID << ";\n";
				if(speech.responses[i].link != 0)
					out << "\t\t\tnextpage = " << static_cast<int> (abs(static_cast<double> (speech.responses[i].link))) << ";\n\t\t\tclosedialog = " << (speech.responses[i].link >= 0 ? "true" : "false" ) << ";\n";
				out << "\t\t}\n";
			}
		}
		out << "\t}\n";
	}
	if(lastID != -1)
		out << "}\n";
}
