//------------------------------------------------------------------------------
// main.cpp
//------------------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//------------------------------------------------------------------------------
// Copyright (C) 2010 "Blzut3" <admin@maniacsvault.net>
//------------------------------------------------------------------------------

#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdlib>
#include <cstdarg>

#include "compile.h"

using namespace std;

enum Mode
{
	MODE_COMPILE,
	MODE_COMPILEOLD,
	MODE_DECOMPILE,
	MODE_DECOMPILEOLD
};

void Error(const char* message, ...)
{
#ifdef unix
	cerr << "\033[1m\033[31mError: ";
#else
	cerr << "Error: ";
#endif

	va_list list;
	va_start(list, message);
	vfprintf(stderr, message, list);
	va_end(list);

#ifdef unix
	cerr << "\033[0m\n";
#else
	cerr << "\n";
#endif
	exit(1);
}

void Warning(const char* message, ...)
{
#ifdef unix
	cerr << "\033[1m\033[33mWarning: ";
#else
	cerr << "Warning: ";
#endif

	va_list list;
	va_start(list, message);
	vfprintf(stderr, message, list);
	va_end(list);

#ifdef unix
	cerr << "\033[0m\n";
#else
	cerr << "\n";
#endif
}

int main(int argc, char* argv[])
{
	if(argc <= 1)
	{
		cout << "Universal Strife Dialog Compiler                                    Version 1.0\n"
		     << "Copyright 2010 \"Blzut3\"\nDistributed under the terms of the GPL v2\n"
		     << "_______________________________________________________________________________\n\n"
		     << "Usage: usdc [options] infile [-o outfile]\n"
		     << "    -c        Compile the input file. (DEFAULT)\n"
		     << "    -t        Compile in old teaser format, not recommend.\n"
		     << "    -d        Decompile the input file.\n"
		     << "    -o        Output file.\n";
	}
	else
	{
		Mode mode = MODE_COMPILE;
		char* outfile = NULL;
		char* infile = NULL;
		for(int i = 1;i < argc;i++)
		{
			if(argv[i][0] == '-' && strlen(argv[i]) == 2)
			{
				switch(argv[i][1])
				{
					case 'c':
						mode = MODE_COMPILE;
						break;
					case 't':
						mode = mode == MODE_DECOMPILE ? MODE_DECOMPILEOLD : MODE_COMPILEOLD;
						break;
					case 'd':
						mode = mode == MODE_COMPILEOLD ? MODE_DECOMPILEOLD : MODE_DECOMPILE;
						break;
					case 'o':
						if(i+1 < argc)
						{
							i++;
							outfile = argv[i];
						}
					default:
						break;
				}
			}
			else
				infile = argv[i];
		}
		if(infile == NULL)
			Error("No input file!");

		// Read in the file
		fstream input(infile, ios_base::in);
		if(!input.is_open())
			Error("Could not open input file!");
		input.seekg(0, ios_base::end);
		int size = input.tellg();
		char *buffer = new char[size];
		input.seekg(0, ios_base::beg);
		input.read(buffer, size);
		if(input.fail() && !input.eof())
		{
			delete[] buffer;
			Error("Could not read input file!");
		}
		input.close();

		fstream outfileStream;
		ostream *output = &cout;
		if(outfile != NULL)
		{
			outfileStream.open(outfile, ios_base::out | ios_base::binary | ios_base::trunc);
			if(!outfileStream.is_open())
			{
				delete[] buffer;
				Error("Could not open the output file for writing.");
			}
			else
				output = &outfileStream;
		}

		switch(mode)
		{
			default:
				Compile(buffer, size, *output, mode == MODE_COMPILEOLD);
				break;
			case MODE_DECOMPILE:
			case MODE_DECOMPILEOLD:
				Decompile(buffer, size, *output, mode == MODE_DECOMPILEOLD);
				break;
		}

		delete[] buffer;
	}
	return 0;
}
